#include "common.h"

extern "C"
{
#include <errno.h>
#include <fcntl.h>
#include <grp.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <termios.h>
#include <dirent.h>
#include <libgen.h>
#include <stdlib.h>
}

bool gCmdLineActive;
static vector_obj* gCmdLine;
static int gCLCursor;
static int gCursor;

static vector_obj* gPrograms;
vector_obj* gCCandidate;

static int gTCur;
vector_obj* gTty;

#if defined(UNIX98_PTY) || defined(__CYGWIN__)

extern "C"
{
extern int grantpt (int __fd);
extern int unlockpt (int __fd);
extern char *ptsname (int __fd);
}

static int get_slave_pty(int fd, char* name)
{
   strcpy(name, "/dev/pty");

	int slave = -1;

	/* get slave name */
	name = ptsname(fd);
	if(name == NULL) {
		perror("pty:@ptsname()");
		return slave;
	}

	/* open slave pty */
	if((slave = open(name, O_RDWR)) < 0) {
		perror("pty:@open()");
		close(slave);
		return slave;
	}

	return slave;
}
#else
static int get_slave_pty(char* name)
{
    struct group* gptr;
    gid_t gid;
    int slave = -1;

    if((gptr = getgrnam("tty")) != 0) {
        gid = gptr->gr_gid;
    }
    else {
        gid = -1;
    }

    chown(name, getuid(), gid);

    chmod(name, S_IRUSR|S_IWUSR);

    slave = open(name, O_RDWR);

    return slave;
}
#endif

static int get_master_pty(RESULT char name[])
{
    int master = -1;
    int slave_tmp = -1;

#if defined(UNIX98_PTY) || defined(__CYGWIN__)

   strcpy(name, "/dev/ptmx");

	if((master = open(name, O_RDWR)) < 0) {
		perror("ptmx:open()");
		return master;
	}

	/* chown the slave to the user */
	if(grantpt(master) < 0) {
		perror("ptmx:grantpt()");
		close(master);
		return master;
	}

	/* release an internal lock */
	if(unlockpt(master) < 0) {
		perror("ptmx:unlockpt()");
		close(master);
		return master;
	}
#else
    int i, j;

    strcpy(name, "/dev/ptyXX");

    for(i=0; i<16; i++) {
        for(j=0; j<16; j++) {
            name[8] = "pqrstuvwxyzPQRST"[i];
            name[9] = "0123456789abcdef"[j];

            master = open(name, O_RDWR);
            if(master < 0) {
                if(errno == ENOENT) {
                    return (master);
                }
            }
//            else if(master == 0) {
//            }
            else {
                name[5] = 't';

                struct stat statbuf;
                stat(name, &statbuf);
                
                name[5] = 'p';
                
                if(statbuf.st_mode&S_IROTH && statbuf.st_mode&S_IWOTH) {
                    goto forend;
                }

                close(master);
            }
        }
    }
forend:
    
    if((master<0) && (i==16) && (j==16)){
        return master;
    }

    name[5] = 't';
#endif
    return master;
}

void sigwinch_handler(int signal)
{
TBEGIN();

    sMasterTty* current = (sMasterTty*)vector_item(gTty, gTCur);

    struct winsize ws;
    if(ioctl(STDOUT_FILENO, TIOCGWINSZ, &ws) < 0) {
        perror("ptypair: could not get window size");
    }
    if(ioctl(current->mFD, TIOCSWINSZ, &ws) < 0) {
        perror("could not restore window size");
    }
    
TEND();
}

static BOOL sort_name(void* left, void* right)
{
    return strcmp((char*)left, (char*)right) < 0;
}

void cmdline_init()
{
TBEGIN();
    gCmdLine = vector_new(10);
    gPrograms = vector_new(10);
    gCCandidate = vector_new(10);

    gTty = vector_new(10);

    gCmdLineActive = false;
    gCLCursor = 0;
    gCursor = 0;
    
    char* path = getenv("PATH");
    if(path == NULL) {
        fprintf(stderr, "$PATH is NULL\n");
        exit(1);
    }

    char* p = path;
    char buf[512];
    char* p2 = buf;
    
    while(*p) {
        if(*p != ':') {
            *p2++ = *p;
        }
        else {
            *p2 = 0;

            DIR* dir = opendir(buf);
            if(dir) {
                struct dirent* entry;
                while(entry = readdir(dir)) {
                    char path2[512];
                    sprintf(path2, "%s/%s", buf, entry->d_name);

                    struct stat stat_;
                    memset(&stat_, 0, sizeof(struct stat));
                    stat(path2, &stat_);
                    
                    if(strcmp(entry->d_name, ".") != 0
                        && strcmp(entry->d_name, "..") != 0
                        &&
                        (stat_.st_mode & S_IXUSR
                            ||stat_.st_mode & S_IXGRP
                            ||stat_.st_mode & S_IXOTH))
                        {
                        vector_add(gPrograms, STRDUP(entry->d_name));
                        }
                }

                closedir(dir);
            }
            
            p2 = buf;
        }
        
        p++;
    }
    
    *p2 = 0;

    DIR* dir = opendir(buf);
    if(dir) {
        struct dirent* entry;
        while(entry = readdir(dir)) {
            if(strcmp(entry->d_name, ".") != 0
                && strcmp(entry->d_name, "..") != 0)
            {
               vector_add(gPrograms, STRDUP(entry->d_name));
            }
        }

        closedir(dir);
    }

    vector_sort(gPrograms, sort_name);
    
TEND();    
}

void cmdline_final()
{
    for(int i=0; i<vector_size(gPrograms); i++) {
        FREE(vector_item(gPrograms, i));
    }
    vector_delete(gPrograms);
    
    for(int i=0; i<vector_size(gCCandidate); i++) {
        FREE(vector_item(gCCandidate, i));
    }
    vector_delete(gCCandidate);
    
    for(int i=0; i<vector_size(gTty); i++) {
        sMasterTty* item = (sMasterTty*)vector_item(gTty, i);
        
        kill(item->mPID, SIGKILL);
        fprintf(stdout, "PID %d: killed\n", item->mPID);
        FREE(item);
    }
    
    vector_delete(gTty);

    for(int i=0; i<vector_size(gCmdLine); i++) {
        string_delete((string_obj*)vector_item(gCmdLine,i));
    }
    vector_delete(gCmdLine);
}

static void cmdline_move(int cursor)
{
TBEGIN();

    gCLCursor += cursor;

    if(gCLCursor < 0) gCLCursor = 0;
    if(gCLCursor >= vector_size(gCmdLine))
        gCLCursor = vector_size(gCmdLine) -1;

TEND();    
}

static void cursor_move(int v)
{
TBEGIN();

    gCursor += v;

    if(gCursor < 0) gCursor = 0;
    string_obj* str = (string_obj*)vector_item(gCmdLine, gCLCursor);
    if(gCursor > string_length(str)) gCursor = string_length(str);

TEND();    
}

// !! out must have enable allocated memory space
static void extend_macro(char* in, RESULT char* out, RESULT bool* ruby_mode, RESULT bool* output, RESULT bool* quick, RESULT bool* shell_mode,  bool flg)
{
TBEGIN();

    sFile* cursor = ActiveDir()->CursorFile();
    *ruby_mode = false;
    *output = true;
    *shell_mode = false;
    *quick = false;
    
    char* p = in;
    char* cp = out;
    while(*p) {
        if(*p != '%') {
            *cp = *p;
            cp++;
            p++;
        }
        else {
            p++;

            switch(*p) {
            case 'r':
                p++;

                if(flg) {
                    *ruby_mode = true;
                }
                else {
                    *cp++ = '%';
                    *cp++ = 'r';
                }
                break;

            case 'q':
                p++;

                if(flg) {
                    *output = false;
                }
                else {
                    *cp++ = '%';
                    *cp++ = 'q';
                }
                break;

            case 'Q':
                p++;

                if(flg) {
                    *quick = true;
                }
                else {
                    *cp++ = '%';
                    *cp++ = 'Q';
                }
                break;

            case 's':
                p++;

                if(flg) {
                    *shell_mode = true;
                }
                else {
                    *cp++ = '%';
                    *cp++ = 's';
                }
                break;

            case 'f': {
                p++;
                    
                const int len = strlen(cursor->mName);
                *cp++ = '\'';
                memcpy(cp, cursor->mName, len);
                cp += len;
                *cp++ = '\'';
                }
                break;

            case 'F': {
                p++;

                *cp++ = '\'';
                const int len = strlen(ActiveDir()->Path());
                memcpy(cp, ActiveDir()->Path(), len);
                cp += len;

                const int len2 = strlen(cursor->mName);
                memcpy(cp, cursor->mName, len2);
                cp += len2;
                
                *cp++ = '\'';
                }
                break;

            case 'x': {
                p++;
                    
                char* extension = extname(cursor->mName);
                const int len = strlen(extension);

                *cp++ = '\'';
                memcpy(cp, extension, len);
                cp += len;
                *cp++ = '\'';

                FREE(extension);
                }
                break;

            case 'X': {
                p++;

                char* nwe = noextname(cursor->mName);
                const int len = strlen(nwe);

                *cp++='\'';
                memcpy(cp, nwe, len);
                cp += len;
                *cp++='\'';
                    
                FREE(nwe);
                }
                break;

            case 'd': {
                p++;

                if(*p == '1') {
                    p++;

                    char* tmp = STRDUP(ActiveDir()->Path());
                    tmp[strlen(tmp)-1] = 0;
                    char* name = basename(tmp);
                    FREE(tmp);
                    const int len = strlen(name);

                    *cp++='\'';
                    memcpy(cp, name, len);
                    cp += len;
                    *cp++='\'';
                }
                else if(*p == '2') {
                    p++;

                    char* tmp = STRDUP(SleepDir()->Path());
                    tmp[strlen(tmp)-1] = 0;
                    char* name = basename(tmp);
                    FREE(tmp);
                    const int len = strlen(name);

                    *cp++='\'';
                    memcpy(cp, name, len);
                    cp += len;
                    *cp++='\'';
                }
                else {
                    cp++;
                    err_msg("illeagal macro char");
                }
                }
                break;

            case 'D': {
                p++;

                if(*p == '1') {
                    p++;

                    const int len = strlen(ActiveDir()->Path());
                    *cp++='\'';
                    memcpy(cp, ActiveDir()->Path(), len-1);
                    cp += len-1;
                    *cp++='\'';
                }
                else if(*p == '2') {
                    p++;
                        
                    const int len = strlen(SleepDir()->Path());

                    *cp++='\'';
                    memcpy(cp, SleepDir()->Path(), len-1);
                    cp += len-1;
                    *cp++='\'';
                }
                else {
                    cp++;
                    err_msg("illeagal macro char");
                }
                    
                }
                break;

            case 'm':
                p++;

                if(*p == '1') {
                    p++;

                    if(flg) {
                        char buf[4096];
                        ActiveDir()->MarkFilesSQuote(buf);      
                        const int len = strlen(buf);

                        memcpy(cp, buf, len);
                        cp += len;
                    }
                    else {
                        *cp++ = '%';
                        *cp++ = 'm';
                        *cp++ = '1';
                    }
                }
                else if(*p == '2') {
                    p++;

                    if(flg) {
                        char buf[4096];
                        SleepDir()->MarkFilesSQuote(buf);      
                        const int len = strlen(buf);
    
                        memcpy(cp, buf, len);
                        cp += len;
                    }
                    else {
                        *cp++ = '%';
                        *cp++ = 'm';
                        *cp++ = '2';
                    }
                }
                else {
                    cp++;
                    err_msg("illeagal macro char");
                }

                break;

            default:
                cp++;
                err_msg("illeagal macro char");
                break;
            }
        }
    }
    *cp = 0;
    
TEND();
}

void cmdline_start(char* cmd, int position)
{
TBEGIN();

    gCmdLineActive = true;

    char cmd2[1024];
    bool ruby_mode;
    bool output;
    bool shell_mode;
    bool quick;
    extend_macro(cmd, cmd2, &ruby_mode, &output, &quick, &shell_mode, false);

    vector_add(gCmdLine, string_new(cmd2));
    gCLCursor = vector_size(gCmdLine)-1;
    if(position > 0)
        gCursor = position-1;
    else
        gCursor = strlen(cmd2) + position;
    
TEND();
}

int cmdline_select_pty(char* title)
{
    int result = 0;

#if !defined(__CYGWIN__)        
    WINDOW* scr = newwin(vector_size(gTty) + 2, 45, 0, 2);

    int cursor = 0;
        
    while(1) {
        wclear(scr);
        wmybox(scr, 0,0, getmaxx(scr), getmaxy(scr));

        wattron(scr ,A_BOLD);
        mvwprintw(scr, 0, 2, title);
        wattroff(scr, A_BOLD);
        
        for(int i=0; i<vector_size(gTty); i++) {
            sMasterTty* master = (sMasterTty*)vector_item(gTty, i);

            if(i == cursor) {
                wattron(scr, A_REVERSE);
                mvwprintw(scr, i+1, 2, "(%d) %-37s", i+1, master->mName);
                wattroff(scr, A_REVERSE);
            }
            else {
               mvwprintw(scr, i+1, 2, "(%d) %s", i+1, master->mName);
            }
        }

        wrefresh(scr);

        int meta;
        int key = mygetch(&meta);
        if(key == 14 || key == KEY_DOWN) {    // CTRL-N
            cursor++;

            if(cursor >= vector_size(gTty)) cursor = 0;
        }
        else if(key == 16 || key == KEY_UP) {    //CTRL-P
            cursor--;

            if(cursor < 0) cursor = vector_size(gTty)-1;
        }
        else if(key == 10 || key == 13) {		// CTRL-J CTRL-M
            result = cursor;
            break;
        }
        else if(key >= '1' && key <= '0'+ vector_size(gTty)) {
            result = key - '1';
            break;
        }
        else {
            result = -1;
            break;
        }
    }

    delwin(scr);
#else
    int cursor = 0;
    
    while(1) {
        mybox(0, 2, 43, vector_size(gTty) + 2);

        myattron(A_BOLD);
        mymvprintw(0,4, title);
        myattroff(A_BOLD);
        
        for(int i=0; i<vector_size(gTty); i++) {
            sMasterTty* master = (sMasterTty*)vector_item(gTty, i);

            if(i == cursor) {
                myattron(A_REVERSE);
                mymvprintw(i+1, 3, "(%d) %-37s", i+1, master->mName);
                myattroff(A_REVERSE);
            }
            else {
               mymvprintw(i+1, 3, "(%d) %-37s", i+1, master->mName);
            }
        }

        myrefresh();

        int meta;
        int key = mygetch(&meta);
        if(key == 14 || key == KEY_DOWN) {    // CTRL-N
            cursor++;

            if(cursor >= vector_size(gTty)) cursor = 0;
        }
        else if(key == 16 || key == KEY_UP) {    //CTRL-P
            cursor--;

            if(cursor < 0) cursor = vector_size(gTty)-1;
        }
        else if(key == 10 || key == 13) {		// CTRL-J CTRL-M
            result = cursor;
            myclear2();
            break;
        }
        else if(key >= '1' && key <= '0'+ vector_size(gTty)) {
            result = key - '1';
            myclear2();
            break;
        }
        else {
            result = -1;
            myclear2();
            break;
        }
    }
#endif

    return result;
}

typedef void (*sighandler_t)(int);

inline void ptty_loop()
{
    struct termios ot, t;
    
    tcgetattr(STDIN_FILENO, &ot);
    t = ot;
    
#ifdef ECHOPRT
    t.c_lflag &= ~(ICANON | ISIG | ECHO | ECHOCTL | ECHOE| ECHOK | ECHOKE
                         | ECHONL | ECHOPRT);
#else
    t.c_lflag &= ~(ICANON | ISIG | ECHO | ECHOCTL | ECHOE| ECHOK | ECHOKE
                  | ECHONL);
#endif

    t.c_iflag |= IGNBRK;
    t.c_iflag &= ~(IXOFF|IXON);
    t.c_cc[VMIN] = 1;
    t.c_cc[VTIME] = 0;
    tcsetattr(STDIN_FILENO, TCSANOW, &t);
    
    sighandler_t old_winch = signal(SIGWINCH, sigwinch_handler);
    sighandler_t old_int = signal(SIGINT, SIG_IGN);
    sighandler_t old_quit = signal(SIGQUIT, SIG_IGN);
    
    raise(SIGWINCH);

    sMasterTty* current = (sMasterTty*)vector_item(gTty, gTCur);
    
    int done = 0;
    do {
        fd_set ready;
        
        FD_ZERO(&ready);
        FD_SET(STDIN_FILENO, &ready);
        FD_SET(current->mFD, &ready);

        select(current->mFD+1, &ready, NULL, NULL, NULL);

        if(FD_ISSET(current->mFD, &ready)) {
            char buf[8192];
            
            int i = read(current->mFD, buf, 8191);
            if(i > 0) {
                if((current->mOutput + current->mOutputTail + i)
                    >=
                   (current->mOutput + kOutputSize))
                {
                    memmove(current->mOutput
                            , current->mOutput + kOutputSize / 2
                            , kOutputSize / 2);
                    current->mOutputTail -= kOutputSize / 2;
                }

                memcpy(current->mOutput+current->mOutputTail, buf, i);
                current->mOutputTail += i;
            
                write(STDOUT_FILENO, buf, i);
            } else {
                done = 2;
            }
        }
        else if(FD_ISSET(STDIN_FILENO, &ready)) {
            char buf[8192];
            
            int i = read(STDIN_FILENO, buf, 8191);

            if(i > 0) {
                int ctrl_z = 0;
            
                for(int j=0; j<i; j++) {
                    if(buf[j] == 26) {            // CTRL-Z
                        done = 1;

                        char buf2[8192];
                        int c = 0;
                        
                        for(int k=0; k<i; k++) {
                            if(buf[k] != 26) {
                                buf2[c++] = buf[k];
                            }
                            else {
                                ctrl_z++;
                            }
                        }

                        memcpy(buf, buf2, i);
                        break;
                    }
                }
                
                write(current->mFD, buf, i-ctrl_z);
            }
            else {
                done = 2;
            }
        }
        
    } while(!done);

    if(done == 2) {
        for(int i=0; i<vector_size(gTty); i++) {
            sMasterTty* item = (sMasterTty*)vector_item(gTty, i);
            if(item->mFD == current->mFD) {
                FREE(item);
                vector_erase(gTty, i);
                break;
            }
        }
    }

    tcsetattr(STDIN_FILENO, TCSANOW, &ot);

    signal(SIGWINCH, old_winch);
    signal(SIGINT, old_int);
    signal(SIGQUIT, old_quit);

    raise(SIGWINCH);
}

static void new_ptty(char* title, char* cmd, bool output)
{
TBEGIN();

    /// new tty ///
    sMasterTty* master = (sMasterTty*)MALLOC(sizeof(sMasterTty));
    strcpy(master->mName, title);
    master->mOutputTail = 0;
    
    vector_add(gTty, master);
    gTCur = vector_size(gTty)-1;

    /// go ///
    char* argv[4];
    char buf[1024];
            
    argv[0] = "/bin/sh";
    argv[1] = "-c";
    argv[2] = buf;
    argv[3] = NULL;

    if(output)
        sprintf(argv[2], "echo \"$ %s\";( %s ); mfiler %d reread; echo \"\nHIT ENTER KEY\"; read TMPTMP", cmd, cmd, getpid());
    else 
        sprintf(argv[2], "( %s ); mfiler %d reread", cmd, getpid());

    char name[256];

    sMasterTty* current = (sMasterTty*)vector_item(gTty, gTCur);
    current->mFD = get_master_pty(name);

    if(current->mFD < 0) {
        err_msg("new_ptty: could not open master pty");;
        return;
    }
    
    struct winsize ws;
    if(ioctl(STDIN_FILENO, TIOCGWINSZ, &ws) < 0) {
        err_msg("new_ptty: could not get window size");
        return;
    }

    int pid = fork();
    if(pid < 0) {
        err_msg("new_ptty: fork false");
        return;
    }

    /// child process ///    
    if(pid == 0) {
        int slave;

#if defined(UNIX98_PTY) || defined(__CYGWIN__)
        if((slave = get_slave_pty(current->mFD, name)) < 0) {
#else
        if((slave = get_slave_pty(name)) < 0) {
#endif
            fprintf(stderr, "new_ptty: could not open slave pty. pleas check permission of %s", name);
            exit(1);
        }

        close(current->mFD);

        if(setsid() < 0) {
            perror("could not set new controlling tty");
        }

#if defined(TIOCSCTTY)
        if(ioctl(slave, TIOCSCTTY, NULL)) {
            perror("could not set new controlling tty");
        }
#endif
   
        dup2(slave, STDIN_FILENO);
        dup2(slave, STDOUT_FILENO);
        dup2(slave, STDERR_FILENO);

        if(slave > 2) {
            close(slave);
        }
    
        if(ioctl(STDOUT_FILENO, TIOCSWINSZ, &ws) < 0) {
            perror("could not restore window size");
        }

        execvp("/bin/sh", argv);

        exit(1);
    }
    
    current->mPID = pid;
    
    ptty_loop();

TEND();    
}

void cmdline_restore_pty(int n)
{
    myendwin();

    if(n < 0 || n >= vector_size(gTty)) return;

    gTCur = n;
    
    myclear();
    myrefresh();
    
    sMasterTty* current = (sMasterTty*)vector_item(gTty, gTCur);
    
    const int m = mygetmaxx() * mygetmaxy() * 3;
    if(current->mOutputTail < m) {
        write(STDOUT_FILENO, current->mOutput, current->mOutputTail);
    }
    else {
        write(STDOUT_FILENO, current->mOutput + current->mOutputTail -m, m);
    }
    
    ptty_loop();

    myinitscr();
    
    myclear();
    view();
    myrefresh();
}

void cmdline_run(char* str, char* title)
{
TBEGIN();
    char cmd[1024];
    bool ruby_mode;
    bool output;
    bool shell_mode;
    bool quick;
    extend_macro(str, cmd, &ruby_mode, &output, &quick, &shell_mode, true);

    char title2[1024];
    bool tmp1;
    bool tmp2;
    bool tmp3;
    bool tmp4;
    extend_macro(title, title2, &tmp1, &tmp2, &tmp3, &tmp4, false);

    char* term = getenv("TERM");
    if(term == NULL) {
        fprintf(stderr, "TERM is null");
        exit(1);
    }

    struct winsize ws;
    ioctl(STDOUT_FILENO,TIOCGWINSZ,&ws);
    const int maxy = ws.ws_row;

    myclear_online(maxy-2, false);
    myclear_online(maxy-1, true);
    
    mymove(maxy-2, 0);
    myrefresh();
    
    /// ruby mode ///
    if(ruby_mode) {
        if(!quick) {
            myendwin();

            //system("clear");
        }

        char buf[256];
        sprintf(buf, "echo \"> %s\"", cmd);
        if(output && !quick) system(buf);
        
        int error;
        rb_eval_string_protect(cmd, &error);

        if(!quick) {
            myinitscr();
        
            if(output) {
                int meta;
                mygetch(&meta);
            }
        
            myclear();
        }
    }
    /// shell mode ///
    else  if(shell_mode || quick || gNoPty) {
        if(!quick) {
            myendwin();
            //system("clear");
        }

        char buf[256];
        sprintf(buf, "echo \"$ %s\"", cmd);
        if(output && !quick) system(buf);
            
        system(cmd);

        if(!quick) {
            myinitscr();
        
            if(output) {
                int meta;
                mygetch(&meta);
            }
        
            myclear();
        }
    }
    /// screen terminal ///
    else if(strstr(term, "screen") == term || gGnuScreen) {
        char* argv[7];

        char buf[4096];
            
        argv[0] = "screen";
        argv[1] = "-t";
        argv[2] = title2;
        argv[3] = "/bin/sh";
        argv[4] = "-c";
        argv[5] = buf;
        argv[6] = NULL;

        if(output && !quick)
            sprintf(argv[5], "echo \"$ %s\";( %s ); mfiler %d reread; echo \"\nHIT ENTER KEY\"; read TMPTMP", cmd, cmd, getpid());
        else 
            sprintf(argv[5], "( %s ); mfiler %d reread", cmd, getpid());

        pid_t pid = fork();
        if(pid == -1) {
            fprintf(stderr, "fork failed");
            exit(1);
        }
        
        if(pid == 0) execvp("screen", argv);
    }

    /// ptty terminal ///
    else {
        if(!quick) {
            myclear();
            myrefresh();
            myendwin();

            system("clear");
        }

        myclear();
        myrefresh();
        new_ptty(title2, cmd, output);

        if(!quick) {
            myinitscr();
        
            myclear();
        }
    }
    
    gLDir->Reread();
    gRDir->Reread();
    
    
TEND();
}

static void completion(vector_obj* all_candidate, char* editing)
{

    for(int i=0; i<vector_size(gCCandidate); i++) {
        FREE(vector_item(gCCandidate, i));
    }
    vector_clear(gCCandidate);
     
    if(strcmp(editing, "") == 0) {
        for(int i=0; i<vector_size(all_candidate); i++) {
            vector_add(gCCandidate
                , STRDUP((char*)vector_item(all_candidate, i)));
        }
    }
    else {
        for(int i=0; i<vector_size(all_candidate); i++) {
            char* item = (char*)vector_item(all_candidate, i);
            if(strstr(item, editing) == item) {
                vector_add(gCCandidate, STRDUP(item));
            }
        }
    }
            
    if(vector_size(gCCandidate) > 0) {
        int l = strlen(editing);
    
        while(l < strlen((char*)vector_item(gCCandidate, 0))) {
            bool same = true;
            
            for(int i= 0; i<vector_size(gCCandidate); i++) {
                char* item = (char*)vector_item(gCCandidate, i);
                
                if(item[l] != ((char*)vector_item(gCCandidate, 0))[l]) {
                    same = false;
                }
            }
    
            if(same) {
                char tmp[2];
                sprintf(tmp, "%c", ((char*)vector_item(gCCandidate, 0))[l]);
                string_insert((string_obj*)vector_item(gCmdLine, gCLCursor)
                                , gCursor, tmp);
                gCursor++;
            }
            else {
                break;
            }
    
            l++;
        }
    }
}

void cmdline_input(int meta, int key)
{
TBEGIN();

    if(key == 6 || key == KEY_RIGHT) {    // CTRL-F
        if(is_kanji(string_c_str((string_obj*)vector_item(gCmdLine,gCLCursor))
                            [gCursor]))
        {
            cursor_move(2);
        }
        else {
            cursor_move(1);
        }

#if defined(__CYGWIN__)
        if(vector_size(gCCandidate) > 0) myclear2();
#endif

        for(int i=0; i<vector_size(gCCandidate); i++) {
            FREE(vector_item(gCCandidate, i));
        }
        vector_clear(gCCandidate);
        
    }
    else if(key == 2 || key == KEY_LEFT) {    // CTRL-B
        if(gCursor > 0) {
            if(gCursor == 1) {
                cursor_move(-1);
            }
            else {
                if(is_kanji(string_c_str((string_obj*)
                            vector_item(gCmdLine,gCLCursor))[gCursor-2]))
                  {
                      cursor_move(-2);
                  }
                else {
                    cursor_move(-1);
                }
            }
        }


#if defined(__CYGWIN__)
        if(vector_size(gCCandidate) > 0) myclear2();
#endif
        
        for(int i=0; i<vector_size(gCCandidate); i++) {
            FREE(vector_item(gCCandidate, i));
        }
        vector_clear(gCCandidate);
        
    }
    else if(key == 1 || key == KEY_HOME) {    // CTRL-A
        cursor_move(-999);

#if defined(__CYGWIN__)
        if(vector_size(gCCandidate) > 0) myclear2();
#endif

        for(int i=0; i<vector_size(gCCandidate); i++) {
            FREE(vector_item(gCCandidate, i));
        }
        vector_clear(gCCandidate);
        
    }
    else if(key == 5 || key == KEY_END) {    // CTRL-E
        cursor_move(999);

#if defined(__CYGWIN__)
        if(vector_size(gCCandidate) > 0) myclear2();
#endif

        for(int i=0; i<vector_size(gCCandidate); i++) {
            FREE(vector_item(gCCandidate, i));
        }
        vector_clear(gCCandidate);
        
    }
    else if(key == 4 || key == KEY_DC) {    // CTRL-D
        string_obj* str = (string_obj*)vector_item(gCmdLine, gCLCursor);
        if(string_length(str) == 0) {
            gCmdLineActive = false;

#if defined(__CYGWIN__)
            myclear2();
#endif
            
            if(gCLCursor != vector_size(gCmdLine)-1) {
                string_obj* str2 = (string_obj*)vector_item(gCmdLine
                                                     ,vector_size(gCmdLine)-1);
                if(string_length(str2) == 0)
                {
                    string_delete(str2);
                    vector_pop_back(gCmdLine);
                }
            }
            string_delete(str);
            vector_erase(gCmdLine, gCLCursor);
        }
        else {
            if(string_length(str) > gCursor) {
                if(is_kanji(string_c_str(str)[gCursor])) {
                    string_erase(str, gCursor, 2);
                }
                else {
                    string_erase(str, gCursor, 1);
                }
            }
        }


#if defined(__CYGWIN__)
        if(vector_size(gCCandidate) > 0) myclear2();
#endif

        
        for(int i=0; i<vector_size(gCCandidate); i++) {
            FREE(vector_item(gCCandidate, i));
        }
        vector_clear(gCCandidate);
        
    }
    else if(key == 8 || key == KEY_BACKSPACE || key == 127) {    // CTRL-H
        if(gCursor > 0) {
            string_obj* str = (string_obj*) vector_item(gCmdLine, gCLCursor);

            if(gCursor == 1) {
                string_erase(str, gCursor-1, 1);
                cursor_move(-1);
            }
            else {
                if(is_kanji(string_c_str(str)[gCursor-2])) {
                    string_erase(str, gCursor-2, 2);
                    
                    cursor_move(-2);
                } else {
                    string_erase(str, gCursor-1, 1);
                    
                    cursor_move(-1);
                }
            }
        }


#if defined(__CYGWIN__)
        if(vector_size(gCCandidate) > 0) myclear2();
#endif

        
        for(int i=0; i<vector_size(gCCandidate); i++) {
            FREE(vector_item(gCCandidate, i));
        }
        vector_clear(gCCandidate);
        
    }
    else if(key == 11) {    // CTRL-K
        string_obj* str = (string_obj*)vector_item(gCmdLine, gCLCursor);
        string_erase(str, gCursor, string_length(str)-gCursor);


#if defined(__CYGWIN__)
        if(vector_size(gCCandidate) > 0) myclear2();
#endif

        
        for(int i=0; i<vector_size(gCCandidate); i++) {
            FREE(vector_item(gCCandidate, i));
        }
        vector_clear(gCCandidate);
    }
    else if(key == 14 || key == KEY_DOWN) {    // CTRL-N
        cmdline_move(1);
        
        string_obj* str = (string_obj*)vector_item(gCmdLine, gCLCursor);
        gCursor = string_length(str);

        
#if defined(__CYGWIN__)
        if(vector_size(gCCandidate) > 0) myclear2();
#endif

        for(int i=0; i<vector_size(gCCandidate); i++) {
            FREE(vector_item(gCCandidate, i));
        }
        vector_clear(gCCandidate);
    }
    else if(key == 16 || key == KEY_UP) {    //CTRL-P
        cmdline_move(-1);
        string_obj* str = (string_obj*)vector_item(gCmdLine, gCLCursor);
        gCursor = string_length(str);

        
#if defined(__CYGWIN__)
        if(vector_size(gCCandidate) > 0) myclear2();
#endif

        for(int i=0; i<vector_size(gCCandidate); i++) {
            FREE(vector_item(gCCandidate, i));
        }
        vector_clear(gCCandidate);
    }
    else if(key == 12) {// CTRL-L
        myclear2();
        gLDir->Reread();
        gRDir->Reread();
    }
    else if(key == 3 || key == 7 || key == 27) { // CTRL-C CTRL-G Escape
        gCmdLineActive = false;

#if defined(__CYGWIN__)
            myclear2();
#endif
        
        string_obj* str = (string_obj*)vector_item(gCmdLine, gCLCursor);
                
        if(gCLCursor != vector_size(gCmdLine)-1) {
            string_obj* str2 = (string_obj*)vector_item(gCmdLine
                                                     ,vector_size(gCmdLine)-1);
            if(string_length(str2) == 0)
            {
                string_delete(str2);
                vector_pop_back(gCmdLine);
            }
        }

        if(string_length(str) == 0) {
            string_delete(str);
            vector_erase(gCmdLine, gCLCursor);
        }


#if defined(__CYGWIN__)
        if(vector_size(gCCandidate) > 0) myclear2();
#endif

        
        for(int i=0; i<vector_size(gCCandidate); i++) {
            FREE(vector_item(gCCandidate, i));
        }
        vector_clear(gCCandidate);
    }
    else if(key == 10 || key == 13) {		// CTRL-J CTRL-M
        string_obj* str = (string_obj*)vector_item(gCmdLine, gCLCursor);
        cmdline_run(string_c_str(str), string_c_str(str));

        /// run history ///
        const int size = vector_size(gCmdLine);
        if(gCLCursor != size-1) {
            string_obj* str = (string_obj*)vector_item(gCmdLine, size-1);
            
            if(string_length(str) == 0) {
                string_delete(str);
                vector_pop_back(gCmdLine);
            }

            if(string_length((string_obj*)vector_item(gCmdLine, gCLCursor))
                    != 0)
                vector_add(gCmdLine, vector_item(gCmdLine, gCLCursor));

            vector_erase(gCmdLine, gCLCursor);
        }
        
        gCmdLineActive = false;


#if defined(__CYGWIN__)
        if(vector_size(gCCandidate) > 0) myclear2();
#endif

        
        for(int i=0; i<vector_size(gCCandidate); i++) {
            FREE(vector_item(gCCandidate, i));
        }
        vector_clear(gCCandidate);
    }
    else if(key == 9) {        // CTRL-I
        char* cmdline = string_c_str(
                            (string_obj*)vector_item(gCmdLine, gCLCursor));

        /// get editing text and position ///
        int i = 0;
        int editing_position = 0;

        char cmd_name[512];
        cmd_name[0] = 0;
        
        char editing[512];
        editing[0] = 0;
        char* e = editing;
        
        while(i < gCursor) {
            if(cmdline[i] == ' ' || cmdline[i] == '\t') {
                i++;

                editing[0] = 0;
                e = editing;
            }
            else if(cmdline[i] == '|' || cmdline[i] == ';' || cmdline[i] == '&') {
                i++;

                editing[0] = 0;
                e = editing;
                
                editing_position = 0;
            }
            else {
                editing[0] = 0;
                e = editing;
                
                while(i < gCursor) {
                    if(cmdline[i]==' ' || cmdline[i]=='\t' || cmdline[i]=='|' || cmdline[i]==';' || cmdline[i]=='&') {
                        editing_position++;
                        break;
                    }
                    *e++ = cmdline[i++];
                }
                *e = 0;

                if(editing_position == 1) {
                    strcpy(cmd_name, editing);
                }
            }
        }

M(("cmdname %s editing %s editing_position %d", cmd_name, editing, editing_position));
        
        /// divide editing into dir part and file part ///
        char editing_dir[512];
        char editing_file[512];

        strcpy(editing_dir, "");
        strcpy(editing_file, "");
        
        e = editing + strlen(editing);
        while(e >= editing) {
            if(*e == '/') {
                memcpy(editing_dir, editing, e - editing + 1);
                editing_dir[e-editing + 1] = 0;
    
                memcpy(editing_file, e + 1, editing + strlen(editing) - e);
                editing_file[editing + strlen(editing) -e] = 0;
                break;
            }
                
            e--;
        }
    
M(("editing %s dir %s editing_file %s", editing, editing_dir, editing_file));        
        if(strcmp(editing_dir, "") == 0) {
            strcpy(editing_file, editing);
        }
        
        if(strcmp(editing_dir, "") != 0) {
            DIR* dir = opendir(editing_dir);
            vector_obj* files = vector_new(10);

            if(dir) {
                struct dirent* entry;
                while(entry = readdir(dir)) {
                    if(strcmp(entry->d_name, ".") != 0
                        && strcmp(entry->d_name, "..") != 0)
                    {
                    char buf[PATH_MAX];
                    
                    struct stat _stat;
                    strcpy(buf, editing_dir);
                    strcat(buf, entry->d_name);
                    stat(buf, &_stat);

                    strcpy(buf, entry->d_name);
                    if(S_ISDIR(_stat.st_mode)) {
                        strcat(buf, "/");
                    }
                    
                    vector_add(files, STRDUP(buf));
                    }
                }

                closedir(dir);

                vector_sort(files, sort_name);
                completion(files, editing_file);

                for(int i=0; i<vector_size(files); i++) {
                    FREE(vector_item(files, i));
                }
                vector_delete(files);
            }
        }
        else {
            /// editing program name ///
            if(editing_position == 0) {
                completion(gPrograms, editing);
            }
            
            /// editing file name ///
            else {
                DIR* dir = opendir(".");
                vector_obj* files = vector_new(10);

                if(dir) {
                    struct dirent* entry;
                    while(entry = readdir(dir)) {
                        if(strcmp(entry->d_name, ".") != 0
                            && strcmp(entry->d_name, "..") != 0)
                        {
                        char buf[PATH_MAX];
                    
                        struct stat _stat;
                        strcpy(buf, "./");
                        strcat(buf, entry->d_name);
                        stat(buf, &_stat);

                        strcpy(buf, entry->d_name);
                        if(S_ISDIR(_stat.st_mode)) {
                            strcat(buf, "/");
                        }
                    
                        vector_add(files, STRDUP(buf));
                        }
                    }

                    closedir(dir);

                    vector_sort(files, sort_name);
                    completion(files, editing_file);

                    for(int i=0; i<vector_size(files); i++) {
                        FREE(vector_item(files, i));
                    }
                    vector_delete(files);
                }
            }
        }
    }
    else if(is_kanji(key)) {
        int meta;
        int key2 = mygetch(&meta);

        char tmp[3];
        sprintf(tmp, "%c%c", (char)key, (char)key2);
        string_obj* str = (string_obj*)vector_item(gCmdLine, gCLCursor);
        string_insert(str, gCursor, tmp);
        gCursor+=2;
    }
    else if(meta == 0 && ' ' <= key && key <= '}') {
        if(key == ' ') {

#if defined(__CYGWIN__)
        if(vector_size(gCCandidate) > 0) myclear2();
#endif

            
            for(int i=0; i<vector_size(gCCandidate); i++) {
                FREE(vector_item(gCCandidate, i));
            }
            
            vector_clear(gCCandidate);
        }
    
        char tmp[2];
        sprintf(tmp, "%c", key);
        string_obj* str = (string_obj*)vector_item(gCmdLine, gCLCursor);
        string_insert(str, gCursor, tmp);
        gCursor++;
    }
    
TEND();
}
  
void cmdline_view()
{
TBEGIN();

    const int maxx = mygetmaxx();
    const int maxy = mygetmaxy();
    
#if defined(__CYGWIN__)
    myclear_online(maxy-2, false);
    myclear_online(maxy-1, true);
#endif

    mymvprintw(maxy - 2, 0, "$ ");
    bool kanji = false;
    bool kanji_zure = false;
    string_obj* str = (string_obj*)vector_item(gCmdLine, gCLCursor);

#if defined(__CYGWIN__)
    mymvprintw(maxy - 2, 2, "%s", string_c_str(str));
    
    for(int i=0; i<string_length(str); i++) {
        char c = string_c_str(str)[i];
        
        if(kanji) {
            kanji = false;
        }
        else if(is_kanji(c)) {
            kanji = true;
        }

        if(i+2 == maxx-1) {
            if(kanji) {
                kanji_zure = true;
            }
        }
    }

    if(kanji_zure) {    
        if(gCursor+2 < maxx-1)
            mymove(maxy -2, 2 + gCursor);
        else
            mymove(maxy-1, 2+ gCursor -maxx+1);
    }
    else {
        if(gCursor+2 < maxx)
            mymove(maxy -2, 2 + gCursor);
        else
            mymove(maxy -1, 2 + gCursor - maxx);
    }

#else
    for(int i=0; i<string_length(str); i++) {
        char c = string_c_str(str)[i];
        
        if(kanji) {
            kanji = false;
        }
        else if(is_kanji(c)) {
            kanji = true;
        }

        if(i+2 < maxx-1) {
            mymvprintw(maxy -2, 2 + i, "%c", c);
        }
        else if(i+2 == maxx-1) {
            if(kanji) {
                mymvprintw(maxy -1, 2 + i - maxx+1, "%c", c);
                kanji_zure = true;
            }
            else {
                mymvprintw(maxy -2, 2 + i, "%c", c);
            }
        }
        else {
            if(kanji_zure) {
                if(2 + i -maxx+1 < maxx-1) {
                    mymvprintw(maxy -1, 2 + i - maxx+1, "%c", c);
                }
            }
            else {
                if(2 + i - maxx < maxx-1) {
                    mymvprintw(maxy -1, 2 + i - maxx, "%c", c);
                }
            }
        }
    }

    if(kanji_zure) {    
        if(gCursor+2 < maxx-1)
            mymove(maxy -2, 2 + gCursor);
        else
            mymove(maxy-1, 2+ gCursor -maxx+1);
    }
    else {
        if(gCursor+2 < maxx)
            mymove(maxy -2, 2 + gCursor);
        else
            mymove(maxy -1, 2 + gCursor - maxx);
    }

#endif
    
TEND();
}

void cmdline_completion_view()
{
TBEGIN();

    const int maxx = mygetmaxx();
    const int maxy = mygetmaxy();

#if defined(__CYGWIN__)
    myclear2();
#endif
    
    int row = 0;
    int line = 0;
    for(int i= 0; i<vector_size(gCCandidate); i++) {
        char* item = (char*)vector_item(gCCandidate, i);
    
        if(row == 4) {
            row = 0;
            line++;
        }

        if(line >= maxy-2) break;
        
        mymvprintw(line, row * (maxx/4), "%s", item);
        row++;
    }

TEND();
}
